/* Copyright 1996, Jan D. Wolter and Steven R. Weiss, All Rights Reserved. */

/* PARTIPATION FILE reader.
 *
 * Mar 15, 1993 - Jan Wolter:  Original bbsread version
 * Dec  2, 1995 - Jan Wolter:  Ansification
 * Jan  2, 1996 - Jan Wolter:  Backtalk version
 */

#include "backtalk.h"

#include <ctype.h>
#include <errno.h>
#include <time.h>
#include <sys/stat.h>

#include "sysdict.h"
#include "user.h"
#include "dict.h"
#include "str.h"
#include "sel.h"
#include "file.h"
#include "ps_part.h"
#include "lock.h"

#ifdef HAVE_LIMITS_H
#include <limits.h>
#else
#define INT_MAX 32767
#endif

#define SECS_PER_DAY (60*60*24)


/* PARTICIPATION FILE - Normally one participation file is kept in memory.
 * It is for the conference in the sysdict variable "conf" and has been
 * read if "part_loaded" is true.
 *
 * "part_modified" is true if any change has been made to the stored
 * participation file.  If so, we should write the new one out before exiting.
 *
 * If "no_part" is true, we were unable to open a participation file for the
 * user (possibly because he is anonymous, possibly because of permission
 * problems) so we will run as if he was just joining for the first time, and
 * not attempt to save any updates.
 *
 * The participation file is stored in a linked list.
 *
 * Note that we hold the participation file open between loading it and
 * writing it, so that our lock on it doesn't go away.
 */

int part_loaded= FALSE;		/* Have we got participation file in memory? */
int part_modified= FALSE;	/* Do we need to write out pflist when done? */
int snooping= FALSE;            /* Is conference open for snooping only? */
int no_part= FALSE;		/* Don't use a participation file */
time_t part_date= -1L;		/* Last mod date - -1=dunno, 0=brandnew */
FILE *part_fp= NULL;
char part_name[BFSZ];		/* fullpath name of current particip. file */
char *alias= NULL;		/* User's full name, as in particip. file */

struct pfe {
    int item_no;		/* Item number */
    int resp_no;		/* Number of responses read including item */
    int forgotten;		/* Is this item forgotten? */
    long date;			/* Time of last reading */
    int delete;			/* Should this entry not be written out? */
    struct pfe *next;		/* Pointer to next entry */
    } *pflist;


void writepart(void);
void freepart(void);
int openpf(int snoop);
int loadpart(int snoop);
void newpart(int);


/* CLOSEPART -- Close the currently open participation file if there is one,
 * writing it out if it has been modified.  If any item has been modified,
 * it is very important that this be called to write the pf out.
 */

void closepart()
{
    if (!part_loaded) return;

    if (part_modified && !snooping)
	writepart();
    else
	freepart();

    if (part_fp != NULL)
    {
	fclose(part_fp);
	part_fp= NULL;
	unlock(part_name);
    }

    pflist= NULL;

    part_loaded= FALSE;
    part_date= -1L;
}


/* REMOVEPART -- Close and delete the participation file of the currently open
 * conference.
 */

void removepart()
{
    if (no_part || !part_loaded) return;

    if (part_fp != NULL)
    {
	fclose(part_fp);
	part_fp= NULL;
	unlock(part_name);
    }

    if (!snooping)
    {
#ifdef PARTUTIL
	partutil("-d",sdict(VAR_CONF),sdict(VAR_ID));
#else
	unlink(part_name);
#endif
    }

    pflist= NULL;
    part_loaded= part_modified= FALSE;
    no_part= TRUE;
    part_date= -1L;
}


/* OPENPART - Make sure the participation file for the current conference has
 * been loaded in memory.  If the argument is true, we will open it in "snoop"
 * mode, in which participation files are read-only.  You don't normally have
 * to call this unless you want to snoop, because calling most any routine
 * that does anything to a participation file does call this.  Returns true if
 * a new participation file was created.
 */

int openpart(int snoop)
{

    if (part_loaded) return 0;
    snooping= snoop;

    if (sdict(VAR_ID)[0] == '\0' || sdict(VAR_HOMEDIR)[0] == '\0')
    {
	no_part= TRUE;
	return 0;
    }
    else
	return loadpart(snoop);
}


/* LOADPART -- Read in a user's participation file for the currnt conference
 * given directory name.  Create an empty one (in memory, so far) if he
 * doesn't have one.  If snoop is true, we don't create a participation
 * file.  Returns true if we created one.
 */

int loadpart(int snoop)
{
char buf[BFSZ], *t;
struct pfe *npf,*lpf=NULL;

    fullpartpath(part_name,sdict(VAR_HOMEDIR),sdict(VAR_PARTICIP));

    /* Open, create, and lock the participation file */
    if (openpf(snoop))
    {
	newpart(!no_part);
	return 1;
    }
    else if (no_part || snoop)
	return 0;

    /* Read and check magic code number */
    buf[0]= '\0';
    if (fgets(buf,BFSZ,part_fp) == NULL || strcmp(buf,"!<pr03>\n"))
    {
	int len;
	if ((len= strlen(buf)) < 8 && !strncmp(buf,"!<pr03>",len))
	{
	    /* Ignore truncated participation file */
	    newpart(FALSE);
	    return 1;
	}
	else
	    /* Bomb on things that can't have been part files */
	    die("Participation file %s has bad magic number", part_name);
    }

    /* Read alias name */
    if (fgets(buf,BFSZ,part_fp) == NULL)
    {
	/* EOF - Assume participation file was truncated */
	newpart(FALSE);
	return 1;
    }
    buf[BFSZ-2]= '\n'; buf[BFSZ-1]= '\0';
    if ((t= strchr(buf,'\n')) == NULL)
    {
	/* No terminating newline - Assume file was truncated */
	newpart(FALSE);
	return 1;
    }
    *t= '\0';
    alias= strdup(buf);
    sets(VAR_ALIAS,buf,SM_COPY);

    while (fgets(buf,BFSZ,part_fp) != NULL)
    {
	if (index(buf,'\n') == NULL)
	    /* Discard lines with no terminating newline */
	    continue;

	npf= (struct pfe *)malloc(sizeof(struct pfe));
#ifdef YAPP_COMPAT
	if (sscanf(buf,"%d %d %lx",&npf->item_no,
				   &npf->resp_no,
				   (unsigned long *)&npf->date) != 3)
	    die("syntax error in participation file %s: %s",part_name,buf);
	if (npf->forgotten= (npf->resp_no < 0))
	    npf->resp_no= -npf->resp_no;
#else
	/* Fancier scanning to handle case of "-0" responses */
	if (!isascii(buf[0]) || !isdigit(buf[0]))
	    die("syntax error in participation file %s: '%s'",part_name,buf);
	npf->item_no= atoi(buf);
	t= firstout(firstin(buf," ")," ");
	if (npf->forgotten= (*t == '-')) t++;
	if (sscanf(t,"%d %lx",&npf->resp_no,
			      (unsigned long *)&npf->date) != 2)
	    die("syntax error in participation file %s: %s",part_name,buf);
#endif /* YAPP_COMPAT */

	npf->delete= 0;
	npf->next= NULL;
	if (lpf == NULL)
	    pflist= npf;
	else
	    lpf->next= npf;
	lpf= npf;
    }

    part_loaded= TRUE;
    part_modified= FALSE;
    return 0;
}


/* NEWPART -- Create a new participation file for the conference in memory.
 * If the "newuser" flag is true, we construct a participation file with some
 * unseen items and some new items, as determined by the "new_isel" and
 * "new_since" settings in the conference settings file.  Otherwise, we
 * just make all items new.
 */

void newpart(int newuser)
{
    pflist= NULL;

    /* Set alias name */
    alias= strdup(user_fname());
    sets(VAR_ALIAS,alias,SM_COPY);

    part_loaded= TRUE;
    part_modified= !no_part;

    if (newuser)
    {
    	char new_isel[BFSZ+1];
    	int new_since, item, nresp, flags;
    	time_t idate, rdate, sdate, now;
	struct pfe *pfn, *pfp;

	part_date= 0L;

	/* Get instructions for which should be new from the settings file  */
	get_settings(sdict(VAR_CONFDIR),new_isel,&new_since);

	/* Everything is new - do nothing */
	if (!strcmp(new_isel,"1-$")) return;

	now= time(0L);
	sdate= now - new_since * SECS_PER_DAY;

	pfn= pfp= NULL;
	for (seek_item(1); !next_item(&item,&nresp,&flags,&idate,&rdate); )
	{
	    pfn= (struct pfe *)malloc(sizeof(struct pfe));
	    pfn->item_no= item;
	    pfn->delete= 0;
	    pfn->forgotten= 0;
	    pfn->resp_no= 0;

	    if ((new_since > 0 && rdate >= sdate) || in_sel(item,new_isel))
		pfn->date= 0;	/* Item New */
	    else
		pfn->date= now;	/* Item Unseen */

	    pfn->next= NULL;
	    if (pfp == NULL)
	    	pflist= pfn;
	    else
		pfp->next= pfn;
	    pfp= pfn;
	}

	/* Special handling for last item and $ sign in isels */
	if (pfn != NULL && pfn->date != 0 && strchr(new_isel,'$') != NULL)
	    pfn->date= 0;
    }
}


/* LAST_READ -- Return the number of the last response to item that the user
 * has read.  Return resp_no = -1 if it is unread.  The flag forgotten is
 * set to true if the item has been forgotten.
 */

void last_read(int item, int *resp_no, int *forgotten, time_t *date)
{
struct pfe *pf;
	
    /* Set default values */
    *forgotten= FALSE;
    *resp_no= -1;
    *date= 0;

    openpart(0);

    if (no_part) return;

    for (pf= pflist; pf != NULL; pf= pf->next)
    {
	/* Item not in participation file...is new */
	if (pf->item_no > item)
	    return;
	
	/* If we found it, return the response number */
	if (pf->item_no == item)
	{
	    *forgotten= pf->forgotten;
	    *resp_no= pf->resp_no-1;
	    *date= pf->date;
	    return;
	}
    }
    return;
}


/* UPDATE_ITEM -- Mark an item read up through the given response.  If the
 * response is -1, the entire item is unseen.  If the response is -2, forget
 * the item.  If response is -3, don't update the response number -- only
 * update the time.  If the response is -4, remember (unforget) the item.
 * If the item is forgotten, marking it seen does not unforget it.
 * This version is fancied up so that multiple calls on the same item in
 * rapid succession can be handled efficiently.  Unless rewind is set, the
 * this will only increase the number of responses seen.  If resp is -1, it
 * will not be changed.
 */

struct pfe *last_part= NULL;

void update_item(int item, int resp, time_t now, int rewind)
{
struct pfe *pf, *pfp, *pfn;
	
    openpart(0);

    if (no_part) return;

    if (last_part != NULL && last_part->item_no == item)
	pf= last_part;
    else
    {
	for (pf= pflist, pfp= NULL; pf != NULL; pfp= pf, pf= pf->next)
	{
	    if (pf->item_no >= item)
		break;
	}

	if (pf == NULL || pf->item_no != item)
	{
	    /* Item not in participation file...put it in */
	    pfn= (struct pfe *)malloc(sizeof(struct pfe));
	    pfn->item_no= item;
	    if (resp == -2)
	    {
	        /* forget an unseen item */
#ifdef YAPP_COMPAT
		pfn->resp_no= 1;
#else
		pfn->resp_no= 0;
#endif
		pfn->forgotten= 1;
		pfn->date= 0;
	    }
	    else
	    {
		pfn->resp_no= (resp < 0) ? 0 : resp+1;
		pfn->forgotten= 0;
		pfn->date= now;
	    }
	    pfn->delete= 0;
	    pfn->next= pf;
	    if (pfp == NULL)
		pflist= pfn;
	    else
		pfp->next= pfn;
	    part_modified= TRUE;
	    return;
	}
    }

    /* Update existing entry */
    if (resp == -2 || resp == -4)
    {
	/* Remembering and Forgetting */
	if ((resp == -4) == (pf->forgotten)) 
	{
#ifdef YAPP_COMPAT
	    /* Yapp participation file format can't handle forgetting unread
	     * items so mark the first response read in that case.
	     */
	    if (pf->resp_no == 0) pf->resp_no= 1;
#endif
	    pf->forgotten= !pf->forgotten;
	    part_modified= TRUE;
	}
    }
    else
    {
	if (resp != -3 && (rewind || pf->resp_no <= resp))
	    pf->resp_no= resp+1;
	pf->date= now;
	part_modified= TRUE;
    }
    last_part= pf;
}


/* LOST_ITEMS -- The items between first and last are missing.  Delete them from
 * the partipation file.  Really just mark them for deletion.  If last is -1,
 * then delete through the end of the conference.
 */

void lost_items(int first,int last)
{
struct pfe *pf;
int state=1;

    if (no_part || last < first) return;

    openpart(0);

    if (no_part) return;

    if (last == -1) last= INT_MAX;
	
    for (pf= pflist; pf != NULL; pf= pf->next)
    {
	if (state == 1 && pf->item_no >= first)
	    state= 2;
		
	if (state == 2)
	{
	    if (pf->item_no > last) break;

#if 0
	    /* If an item was deleted by Picospan, its index file
	     * may still be there.  But the following test means
	     * every user whoever saw the item before it was
	     * deleted would try to delete the index file once.
	     * That's a lot of overhead.  A cron script cleaning
	     * up leftover index files is a better solution.
	     */
	    if (!pf->delete)
		remove_index(sdict(VAR_CONFDIR),pf->item_no);
#endif
	    pf->delete= 1;
	}
    }
}


/* WRITEPART -- Write out the participation file.  A loadpart() or newpart()
 * call must preceed this.  This deallocates the memory that it was stored
 * in as a side effect.
 */

void writepart()
{
struct pfe *pf, *npf;

    if (no_part)
    {
	if (part_loaded) freepart();
	return;
    }

    rewind(part_fp);

    fputs("!<pr03>\n",part_fp);
    fputs(alias,part_fp);
    fputc('\n',part_fp);
    free(alias);

    for (pf= pflist; pf != NULL; pf= npf)
    {
	if (!pf->delete)
	{
#ifdef YAPP_COMPAT
	    fprintf(part_fp,"%d %d %lX\n",
		    pf->item_no,
		    pf->resp_no * (pf->forgotten ? 1 : -1),
		    (unsigned long)pf->date);
#else
	    /* This version allows a "-0" to be printed, which Picospan
	     * understands just fine, but which Yapp considers freakish.
	     */
	    fprintf(part_fp, pf->forgotten ? "%d -%d %lX\n" : "%d %d %lX\n",
		    pf->item_no,
		    pf->resp_no,
		    (unsigned long)pf->date);
#endif /* YAPP_COMPAT */
	}
	npf= pf->next;
	free(pf);
    }

    /* We need this here because the file can get smaller, and we can't
     * destroy and recreate it because we'd lose our lock.
     */

    ftruncate(fileno(part_fp),(size_t)ftell(part_fp));
}


/* FREEPART -- Deallocate the memory we have the participation file stored
 * in.
 */

void freepart()
{
struct pfe *pf, *npf;

    free(alias);

    for (pf= pflist; pf != NULL; pf= npf)
    {
	npf= pf->next;
	free(pf);
    }
}


/* FULLPFPATH -- Construct the full path name of the participation file,
 * given the file name.  It is saved in the global "part_name" variable.
 * Returns a pointer to "part".
 */

char *fullpartpath(char *path, char *dir, char *name)
{
int homelen;

    strcpy(path,dir);
    homelen= strlen(path);
#ifdef ALLOW_CFDIR
    strcpy(path+homelen,"/.cfdir/");
    if (access(path,0))
	strcpy(path+homelen+1,name);
    else
	strcat(path+homelen,name);
#else
    path[homelen]= '/';
    strcpy(path+homelen+1,name);
#endif /* ALLOW_CFDIR */
    return(path);
}


/* OPENPF - Open the current participation file and store the filepointer in
 * the part_fp global variable.  If there was no previous participation file
 * return 1, otherwise return 0;
 */

int openpf(int snoop)
{
    /* Try opening an existing particiption file */
    part_fp= fopen(part_name, snoop ? "r" : "r+");

#ifdef PARTUTIL
    /* If we fail because of permissions, try using partutil to fix them */
    if (part_fp == NULL && errno == EACCES)
    {
	partutil("-p",sdict(VAR_CONF),sdict(VAR_ID));
	part_fp= fopen(part_name, snoop ? "r" : "r+");
    }
#endif

    /* If we opened file, lock it and return */
    if (part_fp != NULL)
    {
	if (!no_part)
	    lock_exclusive(fileno(part_fp),part_name);
	return 0;
    }
    else if (snoop)
	return 1;

    /* If file exist but isn't readable then don't try to create it and
     * just run without one.
     */
    if (errno == EACCES)
    {
	no_part= TRUE;
	return 1;
    }

    /* File doesn't exist - create one */
#ifdef PARTUTIL
    partutil("-c",sdict(VAR_CONF),sdict(VAR_ID));
#endif
    no_part= (no_part || ((part_fp= fopen(part_name,"w+")) == NULL));
    if (!no_part)
        lock_exclusive(fileno(part_fp),part_name);
    return 1;
}


#ifdef PARTUTIL

#include "waittype.h"

/* PARTUTIL - execute the partutil program to create, permit, or destory
 * user's participation file for a conference.  Flag should be "-c" to
 * create, "-p" to permit, or "-d" to destroy a participation file.  A flag
 * of "-l" permits (or creates, if necessary) a .cflist for the user.  In
 * this case conf should be NULL.
 */

int partutil(char *flag, char *conf, char *user)
{
pid_t cpid,wpid;
waittype status;

    if ((cpid= fork()) == 0)
    {
	/* Child process: do the exec */
	if (conf == NULL)
	    execl(PARTUTIL,"partutil",flag,user,(char *)NULL);
	else
	    execl(PARTUTIL,"partutil",flag,conf,user,(char *)NULL);
	exit(127);
    }

    while ((wpid= wait(&status)) != -1 && wpid != cpid)
	;

    return((WIFEXITED(status)) ? WEXITSTATUS(status) : 127);
}
#endif /* PARTUTIL */


/* SET_CONF_ALIAS - Change the default full name for the user for the
 * currently open conference.  "newname" should be malloc'ed memory.
 * Aliases with newlines are clipped at the newline.
 */

void set_conf_alias(char *newname)
{
    char *p;
    if ((p= strchr(newname,'\n')) != NULL) *p= '\0';
    if (alias != NULL) free(alias);
    alias= newname;
    part_modified= TRUE;
}

/* PARTDATE - Return the last modification date of the currently conference's
 * participation file.  If it didn't exist before this session, return 0.
 */

time_t partdate()
{
    struct stat st;
    if (!part_loaded) return 0L;
    if (part_date == -1L)
	part_date= stat(part_name,&st) ? 0L : st.st_mtime;
    return part_date;
}
